/*
 * log.c  -  Logging functions for watchdog
 *
 * Copyright (C) 2008 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id$
 */


#include "compat.h"
#include "watchdog.h"
#ifdef HAVE_SYSLOG_H
# include <syslog.h>
#endif



/*
 ****************************************************************************
 *
 * Local variables
 *
 */
static FILE *logfile = NULL;
static bool initialized = false;



/*
 ****************************************************************************
 *
 * Initialize log file handling. This routine gets called from the watchdog
 * daemon process. stderr is still available.
 *
 */
void log_init(const char *fname)
{
#ifdef HAVE_SYSLOG_H
  int err;
#endif

  /* Check if we have something to do */
  if (initialized)
	return;

  /* Open syslog logging if available */
  if (fname == NULL || !strcmp(fname, "syslog")) {
#ifdef HAVE_SYSLOG_H
	openlog(progname, LOG_PID, LOG_DAEMON);
#endif
	logfile = NULL;
	initialized = true;
	return;
  }

  /* Open a log file. If we are unable, ignore the log file and log to syslog */
  logfile = fopen(fname, "a+");
#ifdef HAVE_SYSLOG_H
  if (logfile == NULL) {
	err = errno;
	openlog(progname, LOG_PID, LOG_DAEMON);
# ifdef HAVE_STRERROR
	syslog(LOG_ERR, "Unable to open log file %s: %s", fname, strerror(err));
# else
	syslog(LOG_ERR, "Unable to open log file %s: Error %d", fname, err);
# endif
	syslog(LOG_NOTICE, "Continuing with syslog logging");
  }
#endif
  initialized = true;
}



/*
 ****************************************************************************
 *
 * Close the log file
 *
 */
void log_close(void)
{
  if (initialized) {
	if (logfile != NULL) {
		fclose(logfile);
		logfile = NULL;
	}
#ifdef HAVE_SYSLOG_H
	else
		closelog();
#endif
	initialized = false;
  }
}



/*
 ****************************************************************************
 *
 * Write a notice message to the log file. If we are not in daemon mode, write
 * to stderr instead.
 *
 */
void log_message(bool iserror, const char *msg)
{
  if (!initialized)
	fprintf(stderr, "%s: %s\n", progname, msg);
  else if (logfile != NULL)
	fprintf(logfile, "%s\n", msg);
#ifdef HAVE_SYSLOG_H
  else if (iserror)
	syslog(LOG_NOTICE, "%s", msg);
  else
	syslog(LOG_ERR, "%s", msg);
#endif
}



/*
 ****************************************************************************
 *
 * Write an error message to the log file. If we are not in daemon mode, write
 * to stderr instead.
 *
 */
void log_error(const char *msg)
{
  if (!initialized)
#ifdef HAVE_STRERROR
	fprintf(stderr, "%s: %s: %s\n", progname, msg, strerror(errno));
#else
	fprintf(stderr, "%s: %s: Error %d\n", progname, msg, errno);
#endif
  else if (logfile != NULL)
#ifdef HAVE_STRERROR
	fprintf(logfile, "%s: %s\n", msg, strerror(errno));
#else
	fprintf(logfile, "%s: Error %d\n", msg, errno);
#endif
#ifdef  HAVE_SYSLOG_H
  else
	syslog(LOG_ERR, "%s: %m", msg);
#endif
}

