/*
 * main.c  -  Main module for watchdog program
 *
 * Copyright (C) 2008 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id$
 */


#include "compat.h"
#include "watchdog.h"
#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif
#ifdef HAVE_SIGNAL_H
# include <signal.h>
#endif
#ifdef HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif
#ifndef _PATH_DEVNULL
# ifdef _PATH_DEV
#  define _PATH_DEVNULL	_PATH_DEV "null"
# else
#  define _PATH_DEVNULL	"/dev/null"
# endif
#endif



/*
 ****************************************************************************
 *
 * Define nice level for daemon
 *
 */
#ifdef HAVE_SETPRIORITY
# ifndef NICELEVEL
#  define NICELEVEL	-10
# endif
#endif



/*
 ****************************************************************************
 *
 * Global variables
 *
 */
char *progname = NULL;
int verbose = 0;



/*
 ****************************************************************************
 *
 * Local variables
 *
 */
static char *portname = PATH_SERIALPORT;
static char *logfile = NULL;
static volatile bool terminate = false;
static bool daemonize = false;
static bool killwd = false;
static bool resetclr = false;
static bool trigger = false;
static bool varget = false;
static bool varset = false;
static bool getstatus = false;
static int varindex = 0;
static int varvalue = 0;



/*
 ****************************************************************************
 *
 * Configuration register definitions
 * These definitions have to appear in the order they are used in the
 * configuration EEPROM of the hardware device.
 *
 */
static struct {
	char  *name;
	int    min;
	int    max;
} conf_regs[] = {
	{ "por_timeout",	5,		 50		},
	{ "power_timeout",	2,		 10		},
	{ "off_timeout",	5,		100		},
	{ "on_timeout",		5,		100		},
	{ "wd_timeout",		WD_TIMEOUT_MIN,	WD_TIMEOUT_MAX	},
	{ "restart_timeout",	20,		255		},
	{ "int_timeout",	0,		100		},
	{ "reset_count",	0,		255		},
	{ NULL,			0,		0		}
};



/*
 ****************************************************************************
 *
 * Convert configuration register name into register index number
 *
 */
static void convreg(char *name, bool doset)
{
  int val = 0;
  int i;

  /* Check if the name is a number or a register name */
  if ((*name >= '0' && *name <= '9') || *name == '-') {
	i = 1;
	while (name[i] >= '0' && name[i] <= '9')
		i++;
	if (name[0] == '-' || (name[i] != '\0' && name[i] != ':') || i > 3) {
		log_message(true, "Invalid configuration index number");
		exit(1);
	}
	if (name[i] == ':') {
		val = i + 1;
		name[i] = '\0';
	}
	i = atoi(name);
	if (i < WD_MIN_CONFIG || i > WD_MAX_CONFIG) {
		log_message(true, "Configuration index number out of range");
		exit(1);
	}
	varindex = i;
  } else {
	i = 0;
	while (name[i] != '\0' && name[i] != ':')
		i++;
	if (name[i] == ':') {
		val = i + 1;
		name[i] = '\0';
	}
	i = 0;
	while (conf_regs[i].name != NULL) {
		if (!strcmp(conf_regs[i].name, name))
			break;
		i++;
	}
	if (conf_regs[i].name == NULL) {
		log_message(true, "Unknown configuration index name");
		exit(1);
	}
	varindex = i;
  }

  /* Check if we need a variable value */
  if (!doset) {
	if (val > 0) {
		log_message(true, "Configuration register value not allowed");
		exit(1);
	}
	return;
  } else if (val == 0) {
	log_message(true, "Missing configuration register value");
	exit(1);
  }

  /* Convert variable value */
  i = val;
  while (name[i] >= '0' && name[i] <= '9')
	i++;
  if (name[i] != '\0' || (i - val) > 3) {
	log_message(true, "Invalid configuration register value");
	exit(1);
  }
  varvalue = atoi(&(name[val]));
  if (varvalue < conf_regs[varindex].min ||
      varvalue > conf_regs[varindex].max) {
	log_message(true, "Configuration register value out of range");
	exit(1);
  }
}



/*
 ****************************************************************************
 *
 * Print usage message
 *
 */
static void usage(void)
{
  int i;

  fprintf(stderr, "Usage: %s -d [-p <port>] [-l <logfile>] [-t] [-v]\n", progname);
  fprintf(stderr, "       %s -g <index> [-p <port>] [-v]\n", progname);
  fprintf(stderr, "       %s -s <index>:<value> [-p <port>] [-v]\n", progname);
  fprintf(stderr, "       %s -r [-p <port>] [-v]\n", progname);
  fprintf(stderr, "       %s -q [-p <port>] [-v]\n", progname);
  fprintf(stderr, "       %s -k [-v]\n", progname);
  fprintf(stderr, "       %s -h\n", progname);
  fprintf(stderr, "       %s -V\n\n", progname);
  fprintf(stderr, "Supported configuration index names:\n");
  for (i = WD_MIN_CONFIG; i <= WD_MAX_CONFIG; i++)
	fprintf(stderr, "  %3d - %s\t[%3d..%3d]\n", i,
			conf_regs[i].name, conf_regs[i].min, conf_regs[i].max);
  exit(1);
}



/*
 ****************************************************************************
 *
 * Handle signals: this terminates the connection to the watchdog and
 * ends the program.
 *
 */
static RETSIGTYPE sighandler(int signum)
{
  terminate = true;
  signal(signum, SIG_IGN);
#if RETSIGTYPE != void
  return((RETSIGTYPE)0);
#endif
}



/*
 ****************************************************************************
 *
 * Run the watchdog daemon. This routine forks and handles the watchdog
 * daemon actions. It never returns.
 *
 */
static void run_daemon(void)
{
  FILE *lock;
  pid_t pid;
  int err;

  /* Daemonize only when not debugging */
  if (verbose < 2) {

	/* Fork off the parent process */
	pid = fork();
	if (pid < 0) {
		log_error("Unable to fork the watchdog daemon");
		wd_close();
		exit(1);
	}

	/*
	 * If we got a PID, then we can exit the parent process. Don't call
	 * wd_close() as that would reset the serial port attributes. The
	 * child process inherits the serial port.
	 */
	if (pid > 0) {
#ifdef HAVE_SETPRIORITY
		if (getuid() == 0)
			setpriority(PRIO_PROCESS, pid, NICELEVEL);
#endif
		exit(0);
	}

	/* Initialize log file handling */
	log_init(logfile);
	log_message(false, "Watchdog daemon " PACKAGE_VERSION " started");

	/* Detach from controlling terminal */
	if (setsid() < 0) {
		log_error("Unable to detach from controlling terminal");
		wd_close();
		exit(1);
	}

	/* Redirect standard files to /dev/null */
	freopen(_PATH_DEVNULL, "r", stdin);
	freopen(_PATH_DEVNULL, "w", stdout);
	freopen(_PATH_DEVNULL, "w", stderr);

	/* Change the current working directory to something benign */
	if (chdir("/tmp") < 0) {
		log_error("Unable to change working directory");
		wd_close();
		exit(1);
	}
  }

  /* Write the PID to the lock file */
  lock = fopen(PID_FILE, "w+");
  if (lock == NULL) {
	log_error("Unable to create lock file");
	wd_close();
	exit(1);
  }
  fprintf(lock, "%d", getpid());
  fclose(lock);

  /* Set signals to close the serial port on abort */
  signal(SIGINT, &sighandler);
  signal(SIGQUIT, &sighandler);
  signal(SIGTERM, &sighandler);
  signal(SIGALRM, SIG_IGN);
  signal(SIGCHLD, SIG_IGN);
#ifdef SIGHUP
  signal(SIGHUP, SIG_IGN);
#endif
#ifdef SIGIO
  signal(SIGIO, SIG_IGN);
#else
# ifdef SIGPOLL
  signal(SIGPOLL, SIG_IGN);
# endif
#endif
#ifdef SIGUSR1
  signal(SIGUSR1, SIG_IGN);
#endif
#ifdef SIGUSR2
  signal(SIGUSR2, SIG_IGN);
#endif

  /* Now activate the watchdog */
  err = 0;
  if (wd_daemon_on()) {
	log_message(false, "Watchdog timer started");
	while (!terminate) {
		if (!wd_daemon_sleep()) {
			sync();
			err = 1;
			break;
		}
		if (trigger || verbose > 1)
			log_message(false, "Watchdog trigger");
	}
  }

  /* Turn watchdog off before terminating the daemon */
  wd_daemon_off();
  unlink(PID_FILE);
  if (verbose < 2) {
	log_message(false, "Watchdog daemon terminated");
	log_close();
  }
  exit(err);
}



/*
 ****************************************************************************
 *
 * Check if the watchdog is locked
 * Needs to get called without serial port opened!!
 *
 */
static void check_lock(void)
{
  FILE *lock;
  struct stat sb;
  pid_t pid;

  if (stat(PID_FILE, &sb) == 0) {
	lock = fopen(PID_FILE, "r");
	if (lock == NULL) {
		log_error("Watchdog locked, unable to open PID file");
		exit(1);
	}
	if (fscanf(lock, "%d", &pid) != 1) {
		log_error("Watchdog locked, unable to read PID file");
		fclose(lock);
		exit(1);
	}
	fclose(lock);
	if (pid > 1) {
		if (kill(pid, 0) == 0) {
			log_message(true, "Watchdog is locked by daemon");
			exit(1);
		}
		if (errno != ESRCH) {
			log_error("Watchdog locked, unable to find process");
			exit(1);
		}
	}
	if (unlink(PID_FILE) < 0) {
		log_error("Unable to remove stale lock file");
		exit(1);
	}
	log_message(false, "Removed stale lock file");
  }
}



/*
 ****************************************************************************
 *
 * Kill a running watchdog daemon
 * Needs to get called without serial port opened!!
 *
 */
static void do_kill(void)
{
  FILE *lock;
  struct stat sb;
  pid_t pid;
  int i;

  lock = fopen(PID_FILE, "r");
  if (lock == NULL) {
	log_error("Unable to open watchdog PID file");
	exit(1);
  }
  if (fscanf(lock, "%d", &pid) != 1) {
	log_error("Unable to read watchdog PID file");
	fclose(lock);
	exit(1);
  }
  fclose(lock);
  if (pid < 2) {
	log_message(true, "Invalid watchdog PID number");
	exit(1);
  }
  if (kill(pid, SIGQUIT) < 0) {
	log_error("Unable to kill watchdog process");
	exit(1);
  }
  for (i = 0; i < WAIT_KILL; i++) {
	sleep(1);
	if (stat(PID_FILE, &sb) == -1)
		return;
  }
  log_error("Watchdog process not killed");
  exit(1);
}



/*
 ****************************************************************************
 *
 * Main program routine
 *
 */
int main (int argc, char **argv)
{
  char *cp, opt;

  /* Determine our own program name */
  if ((cp = strrchr(argv[0], '/')) == NULL)
	progname = argv[0];
  else
	progname = ++cp;

  /* Process command line options */
  opterr = 0;
  while ((opt = getopt(argc, argv, "Vdg:hkl:p:qrs:tv")) != -1) {
	switch(opt) {
		case 'd':
			daemonize = true;
			break;
		case 'p':
			portname = optarg;
			break;
		case 'q':
			getstatus = true;
			break;
		case 's':
			convreg(optarg, true);
			varset = true;
			break;
		case 'g':
			convreg(optarg, false);
			varget = true;
			break;
		case 'l':
			logfile = optarg;
			break;
		case 'k':
			killwd = true;
			break;
		case 't':
			trigger = true;
			break;
		case 'r':
			resetclr = true;
			break;
		case 'v':
			verbose++;
			break;
		case 'V':
			printf("%s\n", PACKAGE_STRING);
			exit(0);
		case 'h':
		case '?':
		default:
			usage();
	}
  }
  if (optind < argc)
	usage();

  /* Check for plausibility of command line arguments */
  if (!daemonize && !varset && !varget && !resetclr && !killwd && !getstatus) {
	log_message(true, "Need at least one action argument");
	exit(1);
  }
  if ((daemonize + varset + varget + resetclr + killwd + getstatus) != true) {
	log_message(true, "Too many actions specified");
	exit(1);
  }
  if (!daemonize) {
	if (trigger) {
		log_message(true,
			"Trigger option only allowed for daemon operation");
		exit(1);
	}
	if (logfile != NULL) {
		log_message(true,
			"Logfile option only allowed for daemon operation");
		exit(1);
	}
  }
  if (verbose > 1 && logfile != NULL) {
	log_message(true, "Logfile not allowed with debug option");
	exit(1);
  }
#ifndef HAVE_SYSLOG_H
  if (daemonize && (logfile == NULL || !strcmp(logfile, "syslog"))) {
	log_message(true, "Need a logfile name for daemon operation");
	exit(1);
  }
#endif

  /* Kill the watchdog daemon - no need to access the watchdog port */
  if (killwd) {
	do_kill();
	if (verbose > 0)
		log_message(false, "Watchdog daemon terminated");
	exit(0);
  }

  /* Check if the watchdog port is locked before opening it */
  check_lock();

  /* Open the serial port and initialize the watchdog interface */
  wd_open(portname);
  cp = wd_init();
  if (verbose > 0)
	printf("Found watchdog with firmware version number %s\n", cp);

  /* Handle the program action */
  if (varset) {
	/* Set a watchdog variable */
	wd_setval(varindex, varvalue);
	if (verbose > 0)
		printf("Configuration variable %d (%s) set to %d\n",
				varindex, conf_regs[varindex].name, varvalue);
  } else if (varget) {
	/* Print a watchdog variable */
	if (verbose > 0)
		printf("Configuration variable %d (%s) = %d\n",
			varindex, conf_regs[varindex].name, wd_getval(varindex));
	else
		printf("%u\n", wd_getval(varindex));
  } else if (getstatus) {
	/* Get watchdog status */
	if (verbose > 0)
		printf("Watchdog status = %u\n", wd_status());
	else
		printf("%u\n", wd_status());
  } else if (resetclr) {
	/* Clear the watchdog reset counter */
	wd_resetclr();
	if (verbose > 0)
		printf("Watchdog reset counter cleared\n");
  } else if (daemonize) {
	/* Start the watchdog daemon */
	if (verbose > 0)
		printf("Starting watchdog daemon with timeout %u ms\n", wd_setup());
	else
		(void)wd_setup();
	run_daemon();
  }

  /* Finally close the serial port again */
  wd_close();
  return(0);
}

